#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
import unittest
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from parameterized import parameterized

from evo.common.io.interfaces import IDestination, ISource

__all__ = [
    "TEST_DATA",
    "AsyncIterator",
    "TestIDestination",
    "TestISource",
    "asynccontextmanager",
]

TEST_DATA = """
x,y,z
10689.54607,100813.2874,442.172449
10690.87819,100812.6204,437.399555
10692.21062,100811.9502,432.627182
10693.54337,100811.277,427.85533
10694.87638,100810.6009,423.083966
10696.06353,100809.9976,418.831067
10725.47983,100777.2912,448.471682
10727.51038,100775.9784,444.022761
10729.54666,100774.6657,439.576449
10731.58199,100773.3566,435.128615
""".lstrip().encode("utf-8")


class TestISource(unittest.IsolatedAsyncioTestCase):
    """Base class for testing ISource implementations."""

    source: ISource
    """The source to test."""

    def setUp(self) -> None:
        self.setup_source(TEST_DATA)
        self.assertIsInstance(self.source, ISource)

    def setup_source(self, test_data: bytes) -> None:
        """Hook for subclasses to set up the source with the given test data."""
        raise unittest.SkipTest("Subclasses must implement this method.")

    @asynccontextmanager
    async def ctx_test_get_size(self) -> AsyncIterator[None]:
        """Async context manager for testing conditions immediately before and after calling ISource.get_size()"""
        yield

    async def test_get_size(self) -> None:
        """Test that the size is correctly reported."""
        expected_size = len(TEST_DATA)
        async with self.ctx_test_get_size():
            actual_size = await self.source.get_size()
            self.assertEqual(expected_size, actual_size)

    @asynccontextmanager
    async def ctx_test_read_chunk(self, offset: int, length: int) -> AsyncIterator[None]:
        """Async context manager for testing conditions immediately before and after calling ISource.read_chunk().

        offset and length are the arguments that will be passed to read_chunk().

        :param offset: The offset argument that will be passed to read_chunk().
        :param length: The length argument that will be passed to read_chunk().
        """
        yield

    @parameterized.expand(
        [
            ("read first byte", 0, 1),
            ("read last byte", len(TEST_DATA) - 1, 1),
            ("read whole chunk", 0, 256),
            ("read part chunk", 256, 256),
        ]
    )
    async def test_read_chunk(self, _label: str, offset: int, length: int) -> None:
        """Test that the correct data is read from the source."""
        expected_result = TEST_DATA[offset : min(offset + length, len(TEST_DATA))]
        async with self.ctx_test_read_chunk(offset, length):
            actual_result = await self.source.read_chunk(offset, length)
            self.assertEqual(expected_result, actual_result)


class TestIDestination(unittest.IsolatedAsyncioTestCase):
    """Base class for testing IDestination implementations."""

    destination: IDestination
    """The destination to test."""

    def setUp(self) -> None:
        self.setup_destination()
        self.assertIsInstance(self.destination, IDestination)

    def setup_destination(self) -> None:
        """Hook for subclasses to set up the destination."""
        raise unittest.SkipTest("Subclasses must implement this method.")

    @asynccontextmanager
    async def ctx_test_write_chunk(self, offset: int, data: bytes) -> AsyncIterator[None]:
        """Async context manager for testing conditions immediately before and after calling ISource.read_chunk().

        offset and length are the arguments that will be passed to read_chunk().

        :param offset: The offset argument that will be passed to read_chunk().
        :param data: The data argument that will be passed to read_chunk().
        """
        yield

    @parameterized.expand(
        [
            ("write first byte", 0, 1),
            ("write last byte", len(TEST_DATA) - 1, 1),
            ("write whole chunk", 0, 256),
            ("write part chunk", 256, 256),
        ]
    )
    async def test_write_chunk(self, _label: str, offset: int, length: int) -> None:
        """Test that the correct data is written to the destination."""
        data = TEST_DATA[offset : min(offset + length, len(TEST_DATA))]
        async with self.ctx_test_write_chunk(offset, data):
            await self.destination.write_chunk(offset, data)

    async def write_whole_file(self) -> list[tuple[int, bytes]]:
        """Test writing the entire test data contents to the destination.

        :return: A list of tuples of the form (offset, data) where data is the data written at the given byte offset.
        """
        total_bytes = len(TEST_DATA)
        chunks_with_offset = [
            (offset, TEST_DATA[offset : min(offset + 32, total_bytes)])
            for offset in range(0, total_bytes, 32)  # Use really small chunks for the sake of this test.
        ]
        tasks = [self.destination.write_chunk(offset, chunk) for offset, chunk in chunks_with_offset]
        await asyncio.gather(*tasks)
        return chunks_with_offset



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import hashlib
import random
import unittest
import uuid
from collections.abc import Iterator
from datetime import datetime, timezone
from io import BytesIO
from pathlib import Path
from typing import BinaryIO
from unittest import mock
from uuid import UUID

import numpy
import pyarrow as pa
import pyarrow.parquet as pq
from parameterized import parameterized, parameterized_class

from evo.common import Environment
from evo.common.test_tools import BASE_URL
from evo.common.utils import Cache
from evo.objects.exceptions import SchemaValidationError, TableFormatError
from evo.objects.utils import (
    ArrowTableFormat,
    BaseTableFormat,
    KnownTableFormat,
    all_known_formats,
    get_known_format,
)
from evo.objects.utils.tables import _ColumnFormat

SAMPLE_DATA_LENGTH = 10
ENVIRONMENT = Environment(hub_url=BASE_URL, org_id=UUID(int=0), workspace_id=UUID(int=0))
CACHE: Cache


def tearDownModule() -> None:
    """Fail-safe cleanup of the cache directory."""
    CACHE.clear_cache()
    CACHE.root.rmdir()


def setUpModule() -> None:
    cache_dir = Path(__file__).parent.resolve() / f".{__name__.lower()}_cache"
    global CACHE
    CACHE = Cache(cache_dir, mkdir=True)
    unittest.addModuleCleanup(tearDownModule)

    # Write gitignore file in case cleanup fails.
    gitignore_file = cache_dir / ".gitignore"
    gitignore_file.write_text("*\n", encoding="utf-8")


def _all_known_formats_for_testing() -> Iterator[dict]:
    for known_format in all_known_formats:
        yield {"data_format": known_format, "expected_field_names": known_format._field_names}


def _generate_float64_data(n_samples: int) -> Iterator[float]:
    max_ = numpy.finfo("float64").max
    for _ in range(n_samples):
        yield max_ * random.uniform(-1.0, 1.0)


def _generate_int_data(int_type: str, n_samples: int) -> Iterator[int]:
    min_, max_ = numpy.iinfo(int_type).min, numpy.iinfo(int_type).max
    for _ in range(n_samples):
        yield random.randint(min_, max_)


def _generate_bool_data(n_samples: int) -> Iterator[bool]:
    for _ in range(n_samples):
        yield random.choice((True, False))


def _generate_string_data(n_samples: int) -> Iterator[str]:
    str_sample = "0123456789ABCDEF "
    for _ in range(n_samples):
        length = random.randint(10, 10000)
        yield "".join(random.choices(str_sample, k=length))


def _generate_timestamp_data(n_samples: int) -> Iterator[datetime]:
    min_ = datetime(1970, 1, 1, tzinfo=timezone.utc).timestamp()
    max_ = datetime(2038, 12, 31, 23, 59, 59, 999999, tzinfo=timezone.utc).timestamp()
    for _ in range(n_samples):
        yield datetime.utcfromtimestamp(random.uniform(min_, max_))


def _generate_data(format_id: str, n_samples: int) -> Iterator:
    match format_id:
        case "float64":
            yield from _generate_float64_data(n_samples)
        case "uint8" | "uint32" | "uint64" | "int32" | "int64" as int_type:
            yield from _generate_int_data(int_type, n_samples)
        case "bool":
            yield from _generate_bool_data(n_samples)
        case "string":
            yield from _generate_string_data(n_samples)
        case "timestamp":
            yield from _generate_timestamp_data(n_samples)
        case unknown_format:
            raise TypeError(f"Unsupported format '{unknown_format}'")


def _get_sample_column(column_format: _ColumnFormat, n_samples: int) -> pa.Array:
    return pa.array(_generate_data(column_format.id, n_samples), type=column_format.type, size=n_samples)


def _get_table_schema(columns: list[_ColumnFormat]) -> pa.Schema:
    return pa.schema([pa.field(f"{column.id}[{i}]", column.type, nullable=False) for i, column in enumerate(columns)])


def _change_format(current_format: _ColumnFormat) -> _ColumnFormat:
    match current_format.id:
        case "float64":
            return _ColumnFormat("int64")
        case "uint8" | "uint32" | "uint64" | "int32" | "int64":
            return _ColumnFormat("float64")
        case "bool" | "timestamp":
            return _ColumnFormat("string")
        case "string":
            return _ColumnFormat("bool")
        case unknown_format:
            raise TypeError(f"Unsupported format '{unknown_format}'")


def _get_sample_table(
    table_format: BaseTableFormat, n_rows: int, add_column: bool = False, change_types: bool = False
) -> pa.Table:
    column_formats = [column for column in table_format._columns]

    if add_column:
        column_formats.append(_ColumnFormat(column_formats[-1].type))

    if change_types:
        column_formats = [_change_format(column) for column in column_formats]

    if table_format._multi_dimensional:
        # Test multidimensional tables with an arbitrary number of columns. If the number of columns matches a more
        # specific GO type (one with a fixed number of columns), the more specific type would be instantiated.
        column_formats *= 20

    sample_schema = pa.schema(
        [pa.field(f"{column.id}[{i}]", column.type, nullable=False) for i, column in enumerate(column_formats)]
    )
    sample_data = [
        pa.array(_generate_data(column_format.id, n_rows), type=column_format.type, size=n_rows)
        for column_format in column_formats
    ]
    return pa.table(sample_data, names=sample_schema.names).cast(sample_schema)


def _get_buffer_digest(buffer: BinaryIO) -> str:
    """Return a sha256 digest of a binary buffer"""
    buffer.seek(0)
    file_hash = hashlib.sha256()
    while next_block := buffer.read(4 * 1024 * 1024):
        file_hash.update(next_block)
    return file_hash.hexdigest()


def _get_table_digest(table: pa.Table) -> str:
    """return a sha256 digest of a pyarrow.Table saved in parquet format"""
    buffer = BytesIO()
    pq.write_table(table=table, where=buffer, version="2.4", compression="gzip")
    return _get_buffer_digest(buffer=buffer)


def _test_name_from_known_format(cls: type, num: int, params_dict: dict) -> str:
    """Create parameterized test name from DataFormat"""
    data_format = params_dict["data_format"]
    return parameterized.to_safe_name(f"{cls.__name__}_{num}_{data_format.name}")


@parameterized_class(_all_known_formats_for_testing(), class_name_func=_test_name_from_known_format)
class TestKnownFormat(unittest.TestCase):
    data_format: KnownTableFormat
    expected_field_names: str

    def setUp(self) -> None:
        self.sample_table = _get_sample_table(table_format=self.data_format, n_rows=SAMPLE_DATA_LENGTH)
        self.expected_parquet_digest = _get_table_digest(self.sample_table)
        self.data_dir = CACHE.get_location(ENVIRONMENT, self.__class__.__name__)
        self.parquet_file = self.data_dir / self.expected_parquet_digest

    def tearDown(self) -> None:
        self.parquet_file.unlink(missing_ok=True)

    def test_unknown_format(self) -> None:
        all_other_formats = [other for other in all_known_formats if not other.is_provided_by(self.data_format)]
        with (
            mock.patch("evo.objects.utils.table_formats.all_known_formats", new=all_other_formats),
            self.assertRaises(TableFormatError),
        ):
            known_format = get_known_format(self.sample_table)
            raise AssertionError(f"Got unexpected format: {known_format.name}")

    def test_get_known_format(self) -> None:
        actual_format = get_known_format(self.sample_table)
        actual_field_names = actual_format._field_names

        # Check that the actual format is compatible with the expected format.
        self.data_format._check_format(actual_format)

        self.assertEqual(self.data_format.name, actual_format.name)
        self.assertEqual(self.data_format.width, actual_format.width)
        self.assertEqual(self.data_format.data_type, actual_format.data_type)
        self.assertEqual(self.expected_field_names, actual_field_names)

    def test_save_table(self) -> None:
        self.assertFalse(self.parquet_file.is_file())

        known_format = get_known_format(self.sample_table)
        table_info = known_format.save_table(self.sample_table, self.data_dir)

        self.assertEqual(frozenset(table_info), self.expected_field_names)
        self.assertEqual(self.expected_parquet_digest, table_info["data"])
        self.assertEqual(SAMPLE_DATA_LENGTH, table_info["length"])

        self.assertTrue(self.parquet_file.is_file())

    def test_save_table_when_file_exists_and_is_open(self) -> None:
        self.assertFalse(self.parquet_file.is_file())

        known_format = get_known_format(self.sample_table)
        known_format.save_table(self.sample_table, self.data_dir)
        self.assertTrue(self.parquet_file.is_file())

        with self.parquet_file.open():
            table_info = known_format.save_table(self.sample_table, self.data_dir)

        self.assertEqual(frozenset(table_info), self.expected_field_names)
        self.assertEqual(self.expected_parquet_digest, table_info["data"])
        self.assertEqual(SAMPLE_DATA_LENGTH, table_info["length"])

        self.assertTrue(self.parquet_file.is_file())

    def _save_parquet_file(self, add_column: bool = False, add_row: bool = False, change_type: bool = False) -> dict:
        if add_column or add_row or change_type:
            sample_length = SAMPLE_DATA_LENGTH
            if add_row:
                sample_length += 1
            self.sample_table = _get_sample_table(
                table_format=self.data_format, n_rows=sample_length, add_column=add_column, change_types=change_type
            )

        pq.write_table(self.sample_table, where=self.parquet_file, version="2.4", compression="gzip")
        info_dict = {
            "data": self.expected_parquet_digest,
            "length": SAMPLE_DATA_LENGTH,
            "width": self.sample_table.num_columns,
            "data_type": self.data_format.data_type,
            "keys_data_type": self.data_format._columns[0]._format_id,
            "field_names": self.expected_field_names,
        }
        if add_column:
            info_dict["width"] -= 1
        return info_dict

    def test_known_format_from_table_info(self) -> None:
        table_info = self._save_parquet_file()
        self.assertTrue(self.parquet_file.is_file())

        inferred_format = KnownTableFormat.from_table_info(table_info)
        self.assertTrue(self.data_format.is_provided_by(inferred_format))

    def test_load_table(self) -> None:
        table_info = self._save_parquet_file()
        self.assertTrue(self.parquet_file.is_file())
        inferred_format = KnownTableFormat.from_table_info(table_info)

        actual_table = KnownTableFormat.load_table(table_info, self.data_dir)
        self.assertEqual(inferred_format.width, actual_table.num_columns)
        self.assertEqual(table_info["length"], actual_table.num_rows)

        actual_format = ArrowTableFormat.from_schema(actual_table.schema)
        self.assertEqual(inferred_format.data_type, actual_format.data_type)

    def test_load_table_with_too_many_columns(self) -> None:
        table_info = self._save_parquet_file(add_column=True)
        self.assertTrue(self.parquet_file.is_file())

        with self.assertRaises(TableFormatError):
            KnownTableFormat.load_table(table_info, self.data_dir)

    def test_load_table_with_too_many_rows(self) -> None:
        table_info = self._save_parquet_file(add_row=True)
        self.assertTrue(self.parquet_file.is_file())

        with self.assertRaises(SchemaValidationError):
            KnownTableFormat.load_table(table_info, self.data_dir)

    def test_load_table_with_wrong_data_types(self) -> None:
        table_info = self._save_parquet_file(change_type=True)
        self.assertTrue(self.parquet_file.is_file())

        with self.assertRaises(TableFormatError):
            KnownTableFormat.load_table(table_info, self.data_dir)

    def test_load_table_from_uuid(self) -> None:
        table_info = self._save_parquet_file()
        self.assertTrue(self.parquet_file.is_file())

        table_info["data"] = uuid.uuid4()
        self.parquet_file = self.parquet_file.rename(self.parquet_file.parent / str(table_info["data"]))
        self.assertTrue(self.parquet_file.is_file())

        inferred_format = KnownTableFormat.from_table_info(table_info)

        actual_table = KnownTableFormat.load_table(table_info, self.data_dir)
        self.assertEqual(inferred_format.width, actual_table.num_columns)
        self.assertEqual(table_info["length"], actual_table.num_rows)

        actual_format = ArrowTableFormat.from_schema(actual_table.schema)
        self.assertEqual(inferred_format.data_type, actual_format.data_type)


def _complex_formats_for_testing() -> Iterator[dict]:
    """Generator of table formats with their generated GO models and keyword arguments for testing.

    :yields: (<table-format>, <expected-field-names>)
    """
    yield {
        "data_format": BaseTableFormat("downhole-collection-location-path", [pa.float64()] * 3),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": False,
    }
    yield {
        "data_format": BaseTableFormat("hexahedrons-indices", [pa.uint64()] * 8),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": True,
    }
    yield {
        "data_format": BaseTableFormat("hexahedrons-vertices", [pa.float64()] * 3),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": False,
    }
    yield {
        "data_format": BaseTableFormat("quadrilaterals-indices", [pa.uint64()] * 4),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": True,
    }
    yield {
        "data_format": BaseTableFormat("quadrilaterals-vertices", [pa.float64()] * 3),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": False,
    }
    yield {
        "data_format": BaseTableFormat("segments-indices", [pa.uint64()] * 2),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": False,
    }
    yield {
        "data_format": BaseTableFormat("segments-vertices", [pa.float64()] * 3),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": False,
    }
    yield {
        "data_format": BaseTableFormat("tetrahedra-indices", [pa.uint64()] * 4),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": True,
    }
    yield {
        "data_format": BaseTableFormat("tetrahedra-vertices", [pa.float64()] * 3),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": False,
    }
    yield {
        "data_format": BaseTableFormat("triangles-indices", [pa.uint64()] * 3),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": False,
    }
    yield {
        "data_format": BaseTableFormat("triangles-vertices", [pa.float64()] * 3),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": False,
    }
    yield {
        "data_format": BaseTableFormat("unstructured-grid-geometry-vertices", [pa.float64()] * 3),
        "expected_field_names": ["data", "length", "width", "data_type"],
        "expect_extra_column_fails": False,
    }


@parameterized_class(_complex_formats_for_testing(), class_name_func=_test_name_from_known_format)
class TestComplexFormats(unittest.TestCase):
    data_format: BaseTableFormat
    expected_field_names: str
    expect_extra_column_fails: bool

    def setUp(self) -> None:
        self.sample_table = _get_sample_table(table_format=self.data_format, n_rows=SAMPLE_DATA_LENGTH)
        self.expected_parquet_digest = _get_table_digest(self.sample_table)
        self.data_dir = CACHE.get_location(ENVIRONMENT, self.__class__.__name__)
        self.parquet_file = self.data_dir / self.expected_parquet_digest

    def tearDown(self) -> None:
        self.parquet_file.unlink(missing_ok=True)

    def test_save_table(self) -> None:
        self.assertFalse(self.parquet_file.is_file())

        known_format = get_known_format(self.sample_table)
        table_info = known_format.save_table(self.sample_table, self.data_dir)

        self.assertEqual(list(table_info), self.expected_field_names)
        self.assertEqual(self.expected_parquet_digest, table_info["data"])
        self.assertEqual(SAMPLE_DATA_LENGTH, table_info["length"])

        self.assertTrue(self.parquet_file.is_file())

    def test_save_table_extra_column_fails(self) -> None:
        self.assertFalse(self.parquet_file.is_file())
        sample_table = _get_sample_table(self.data_format, n_rows=SAMPLE_DATA_LENGTH, add_column=True)

        if self.expect_extra_column_fails:
            with self.assertRaises(TableFormatError):
                _known_format = get_known_format(sample_table)
        else:
            known_format = get_known_format(sample_table)
            known_format.save_table(sample_table, self.data_dir)

        self.assertFalse(self.parquet_file.is_file())

    def test_save_table_different_column_types_fails(self) -> None:
        self.assertFalse(self.parquet_file.is_file())
        sample_table = _get_sample_table(self.data_format, n_rows=SAMPLE_DATA_LENGTH, change_types=True)

        known_format = get_known_format(sample_table)
        with self.assertRaises(TableFormatError):
            known_format.save_table(self.sample_table, self.data_dir)

        self.assertFalse(self.parquet_file.is_file())



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .client import DownloadedObject, ObjectAPIClient
from .data import ObjectMetadata, ObjectSchema, ObjectVersion, SchemaVersion, Stage
from .io import ObjectDataDownload, ObjectDataUpload

__all__ = [
    "DownloadedObject",
    "ObjectAPIClient",
    "ObjectDataDownload",
    "ObjectDataUpload",
    "ObjectMetadata",
    "ObjectSchema",
    "ObjectVersion",
    "SchemaVersion",
    "Stage",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ._types import DataFrame, Table
from .data import ObjectDataClient

__all__ = [
    "DataFrame",
    "ObjectDataClient",
    "Table",
]

try:
    import pyarrow  # noqa: F401
except ImportError:
    pass  # Omit the following imports if pyarrow is not installed.
else:
    from .table_formats import all_known_formats, get_known_format
    from .tables import ArrowTableFormat, BaseTableFormat, KnownTableFormat

    __all__ += [
        "ArrowTableFormat",
        "BaseTableFormat",
        "KnownTableFormat",
        "all_known_formats",
        "get_known_format",
    ]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from evo.common.io import HTTPSource
from evo.common.test_tools import TestWithDownloadHandler

from .common import TEST_DATA, AsyncIterator, TestISource, asynccontextmanager


class TestHTTPSource(TestISource, TestWithDownloadHandler):
    def setup_source(self, test_data: bytes) -> None:
        TestWithDownloadHandler.setUp(self)
        self.setup_download_handler(test_data)
        self.source = HTTPSource(self.url_generator.get_new_url, self.transport)

    async def test_context_manager(self) -> None:
        """Test context manager functionality"""
        self.transport.assert_no_requests()
        self.transport.open.assert_not_called()
        self.transport.close.assert_not_called()
        self.assertEqual(0, self.url_generator.n_calls)

        with self.assertRaises(AssertionError):
            await self.source.get_size()

        async with self.source:
            self.assertEqual(2, self.transport.open.call_count)
            self.transport.close.assert_called_once()
            self.assertEqual(1, self.url_generator.n_calls)

            # We currently perform a HEAD request to get the size when the context is entered. The headers in the
            # response also tell us whether the server supports range requests.
            self.assert_head_request_made(self.url_generator.current_url)
            self.transport.reset_mock()

        self.transport.assert_no_requests()
        self.transport.close.assert_called_once()
        self.assertEqual(1, self.url_generator.n_calls)

    @asynccontextmanager
    async def ctx_test_get_size(self) -> AsyncIterator[None]:
        self.transport.assert_no_requests()
        async with self.source:
            # We currently perform a HEAD request to get the size when the context is entered.
            self.assert_head_request_made(self.url_generator.current_url)
            self.transport.reset_mock()

            yield

            # The size is cached after the first request.
            self.transport.assert_no_requests()

    @asynccontextmanager
    async def ctx_test_read_chunk(self, offset: int, length: int) -> AsyncIterator[None]:
        self.transport.assert_no_requests()
        async with self.source:
            self.transport.reset_mock()
            yield
            self.assert_range_request_made(self.url_generator.current_url, offset, length)

    async def test_download_file(self) -> None:
        """Test that the source can download a file."""
        test_data_file = self.CACHE_DIR / "test_data_download.csv"
        self.assertFalse(test_data_file.exists())
        await HTTPSource.download_file(str(test_data_file), self.url_generator.get_new_url, self.transport)
        self.assertTrue(test_data_file.exists())
        self.assertEqual(TEST_DATA, test_data_file.read_bytes())


# Delete base test classes to prevent discovery by unittest.
del TestISource



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import enum
import re
from dataclasses import dataclass
from datetime import datetime
from typing import Protocol
from uuid import UUID

from evo.common import ResourceMetadata
from evo.workspaces import ServiceUser

from .exceptions import SchemaIDFormatError

__all__ = [
    "ObjectMetadata",
    "ObjectOrderByEnum",
    "ObjectSchema",
    "ObjectVersion",
    "SchemaVersion",
    "Stage",
]


class ObjectOrderByEnum(str, enum.Enum):
    author = "author"
    created_at = "created_at"
    created_by = "created_by"
    deleted_at = "deleted_at"
    modified_at = "modified_at"
    modified_by = "modified_by"
    object_name = "object_name"


@dataclass(frozen=True, kw_only=True)
class ObjectMetadata(ResourceMetadata):
    """Metadata about a geoscience object."""

    parent: str
    """The parent path of the object."""

    schema_id: ObjectSchema
    """The geoscience object schema."""

    version_id: str
    """An arbitrary identifier for the object version."""

    modified_at: datetime
    """The date and time when the object was last modified."""

    modified_by: ServiceUser | None
    """The user who last modified the object."""

    stage: Stage | None
    """The stage of the object, if available."""

    @property
    def path(self) -> str:
        """The full path of the object, formed by joining the parent and name, separated by a slash ('/')."""
        return f"{self.parent}/{self.name}"

    @property
    def url(self) -> str:
        """The url of the object."""
        return "{hub_url}/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects/{object_id}?version={version_id}".format(
            hub_url=self.environment.hub_url.rstrip("/"),
            org_id=self.environment.org_id,
            workspace_id=self.environment.workspace_id,
            object_id=self.id,
            version_id=self.version_id,
        )


@dataclass(frozen=True, kw_only=True)
class OrgObjectMetadata(ResourceMetadata):
    """Metadata about a geoscience object in an organization listing."""

    workspace_id: UUID
    """The ID of the workspace that contains the object."""

    workspace_name: str | None
    """The name of the workspace that contains the object."""

    schema_id: ObjectSchema
    """The geoscience object schema."""

    modified_at: datetime
    """The date and time when the object was last modified."""

    modified_by: ServiceUser | None
    """The user who last modified the object."""

    stage: Stage | None
    """The stage of the object, if available."""

    @property
    def url(self) -> str:
        """The url of the object."""
        return "{hub_url}/geoscience-object/orgs/{org_id}/workspaces/{workspace_id}/objects/{object_id}".format(
            hub_url=self.environment.hub_url.rstrip("/"),
            org_id=self.environment.org_id,
            workspace_id=self.workspace_id,
            object_id=self.id,
        )


@dataclass(frozen=True, kw_only=True)
class ObjectVersion:
    """Represents a version of an object."""

    version_id: str
    """Used by the service to identify a unique resource version."""

    created_at: datetime
    """A UTC timestamp representing when the version was uploaded to the service."""

    created_by: ServiceUser | None
    """The user that uploaded the version."""

    stage: Stage | None
    """The stage associated with this version, if applicable."""


class _StageModel(Protocol):
    stage_id: UUID
    name: str


@dataclass(frozen=True, kw_only=True)
class Stage:
    """Metadata about a stage"""

    id: UUID
    """The stage UUID."""

    name: str
    """The stage name."""

    @classmethod
    def from_model(cls, model: _StageModel) -> Stage:
        """Create a new instance from an instance of a generated model."""
        return cls(id=model.stage_id, name=model.name)


@dataclass(frozen=True, order=True)
class SchemaVersion:
    major: int
    minor: int
    patch: int

    @classmethod
    def from_str(cls, string: str) -> SchemaVersion:
        return cls(*[int(i) for i in string.split(".")])

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"


@dataclass(frozen=True)
class ObjectSchema:
    root_classification: str
    sub_classification: str
    version: SchemaVersion

    @property
    def classification(self) -> str:
        return f"{self.root_classification}/{self.sub_classification}"

    @classmethod
    def from_id(cls, schema_id: str) -> ObjectSchema:
        schema_components = re.match(
            r"/(?P<root>[-\w]+)/(?P<sub>[-\w]+)/(?P<version>\d+\.\d+\.\d+)/(?P=sub)\.schema\.json",
            schema_id,
        )
        if schema_components is None:
            raise SchemaIDFormatError(f"Could not parse schema id: '{schema_id}'")

        return cls(
            root_classification=schema_components.group("root"),
            sub_classification=schema_components.group("sub"),
            version=SchemaVersion.from_str(schema_components.group("version")),
        )

    def __str__(self) -> str:
        return f"/{self.classification}/{self.version}/{self.sub_classification}.schema.json"



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from pathlib import Path

_THIS_DIR = Path(__file__).parent.resolve()


def _load_json_data(target_file: Path) -> list | dict:
    with target_file.open("r") as json_file:
        return json.load(json_file)


def load_test_data(filename: str) -> list | dict:
    target_file = resolve_file(filename)

    match target_file.suffix.lower():
        case ".json":
            return _load_json_data(target_file)
        case ext:
            raise ValueError(f"Unsupported data file type '{ext}'")


def resolve_file(filename: str) -> Path:
    target_file = (_THIS_DIR / filename).resolve()

    # Test data must live in test data directory.
    if not target_file.is_relative_to(_THIS_DIR):
        raise RuntimeError("Cannot access files outside test data directory.")

    # File must exist.
    if not target_file.exists():
        raise FileNotFoundError(f"Unknown file '{target_file}'")

    return target_file



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import uuid
from unittest import mock

import pyarrow

from evo.blockmodels import BlockModelAPIClient
from evo.blockmodels.endpoints.models import (
    BBox,
    IntRange,
    JobErrorPayload,
    JobResponse,
    JobStatus,
    Mapping,
    QueryDownload,
    QueryResult,
)
from evo.blockmodels.exceptions import CacheNotConfiguredException, JobFailedException
from evo.common import HealthCheckType
from evo.common.data import HTTPHeaderDict, RequestMethod
from evo.common.test_tools import BASE_URL, AbstractTestRequestHandler, MockResponse, TestWithConnector, TestWithStorage
from evo.common.utils import NoFeedback

TABLE = pyarrow.table({"x": [1.0, 2.0], "y": [3.0, 4.0], "z": [5.0, 6.0], "col1": [1, 2], "col2": [3, 4]})
BBOX = BBox(
    i_minmax=IntRange(min=1, max=2),
    j_minmax=IntRange(min=3, max=4),
    k_minmax=IntRange(min=5, max=6),
)


class QueryRequestHandler(AbstractTestRequestHandler):
    def __init__(self, query_result: QueryResult, job_response: JobResponse, pending_request: int = 0) -> None:
        self._query_result = query_result
        self._job_response = job_response
        self._pending_requests = pending_request

    async def request(
        self,
        method: RequestMethod,
        url: str,
        headers: HTTPHeaderDict | None = None,
        post_params: list[tuple[str, str | bytes]] | None = None,
        body: object | str | bytes | None = None,
        request_timeout: int | float | tuple[int | float, int | float] | None = None,
    ) -> MockResponse:
        match method:
            case RequestMethod.POST:
                return MockResponse(status_code=200, content=self._query_result.model_dump_json())
            case RequestMethod.GET:
                if self._pending_requests > 0:
                    self._pending_requests -= 1
                    job_response = JobResponse(job_status=JobStatus.PROCESSING)
                else:
                    job_response = self._job_response
                return MockResponse(status_code=200, content=job_response.model_dump_json())
            case _:
                return self.not_found()


class TestBlockModelAPIClient(TestWithConnector, TestWithStorage):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        TestWithStorage.setUp(self)
        self.bms_client = BlockModelAPIClient(connector=self.connector, environment=self.environment, cache=self.cache)
        self.bms_client_without_cache = BlockModelAPIClient(connector=self.connector, environment=self.environment)

    @property
    def base_path(self) -> str:
        return f"blockmodel/orgs/{self.environment.org_id}/workspaces/{self.environment.workspace_id}"

    async def test_check_service_health(self) -> None:
        """Test service health check implementation"""
        with mock.patch("evo.blockmodels.client.get_service_health", spec_set=True) as mock_get_service_health:
            await self.bms_client.get_service_health()
        mock_get_service_health.assert_called_once_with(self.connector, "blockmodel", check_type=HealthCheckType.FULL)

    async def test_query_block_model(self) -> None:
        bm_uuid = uuid.uuid4()
        version_uuid = uuid.uuid4()
        column_uuid = uuid.uuid4()
        job_uuid = uuid.uuid4()

        download_url = "http://data.com/"

        query_result = QueryResult(
            bm_uuid=bm_uuid,
            version_id=1,
            version_uuid=version_uuid,
            bbox=BBOX,
            mapping=Mapping(columns=[]),
            columns=["col1", str(column_uuid)],
            job_url=BASE_URL + self.base_path + f"blockmodel/{bm_uuid}/jobs/{job_uuid}",
        )
        job_response = JobResponse(
            job_status=JobStatus.COMPLETE,
            payload=QueryDownload(download_url=download_url),
        )
        self.transport.set_request_handler(QueryRequestHandler(query_result, job_response, pending_request=1))

        expected_filename = self.cache.get_location(environment=self.environment, scope="blockmodel") / str(job_uuid)
        with (
            mock.patch("pyarrow.parquet.read_table") as read_table,
            mock.patch("evo.common.io.download.HTTPSource", autospec=True) as mock_source,
        ):

            async def _mock_download_file_side_effect(*args, **kwargs):
                actual_download_url = await kwargs["url_generator"]()
                self.assertEqual(expected_filename, kwargs["filename"])
                self.assertEqual(download_url, actual_download_url)
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIs(NoFeedback, kwargs["fb"])

            mock_source.download_file.side_effect = _mock_download_file_side_effect

            read_table.return_value = TABLE

            result = await self.bms_client.query_block_model_as_table(
                bm_id=bm_uuid,
                columns=["col1", column_uuid],
                bbox=BBOX,
                version_uuid=version_uuid,
            )
        read_table.assert_called_once_with(expected_filename)
        self.assertEqual(TABLE, result)

        # 1 query, 1 pending job status, and 1 complete job status
        self.assertEqual(self.transport.request.call_count, 3)

    async def test_query_block_model_job_failed(self) -> None:
        bm_uuid = uuid.uuid4()
        version_uuid = uuid.uuid4()
        column_uuid = uuid.uuid4()
        job_uuid = uuid.uuid4()

        query_result = QueryResult(
            bm_uuid=bm_uuid,
            version_id=1,
            version_uuid=version_uuid,
            bbox=BBOX,
            mapping=Mapping(columns=[]),
            columns=["col1", str(column_uuid)],
            job_url=BASE_URL + self.base_path + f"blockmodel/{bm_uuid}/jobs/{job_uuid}",
        )
        job_response = JobResponse(
            job_status=JobStatus.FAILED,
            payload=JobErrorPayload(
                detail="Query Job failed",
                status=500,
                title="Query Job failed",
                type="https://seequent.com/error-codes/block-model-service/job/internal-error",
            ),
        )
        self.transport.set_request_handler(QueryRequestHandler(query_result, job_response, pending_request=1))

        with self.assertRaises(JobFailedException):
            await self.bms_client.query_block_model_as_table(
                bm_id=bm_uuid,
                columns=["col1", column_uuid],
                bbox=BBOX,
                version_uuid=version_uuid,
            )

        # 1 query, 1 pending job status, and 1 failed job status
        self.assertEqual(self.transport.request.call_count, 3)

    async def test_query_block_model_no_cache(self) -> None:
        bm_uuid = uuid.uuid4()
        version_uuid = uuid.uuid4()
        column_uuid = uuid.uuid4()

        with self.assertRaises(CacheNotConfiguredException):
            await self.bms_client_without_cache.query_block_model_as_table(
                bm_id=bm_uuid,
                columns=["col1", column_uuid],
                bbox=BBOX,
                version_uuid=version_uuid,
            )

        self.transport.assert_no_requests()
